/*
 * Copyright (c) 2011, 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.jcbf;

import java.awt.Rectangle;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.io.IOException;
import java.nio.ByteOrder;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.imageio.IIOException;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageReaderSpi;
import javax.imageio.stream.ImageInputStream;

/**
 * I read a minimal CBF (miniCBF) image as produced by DECTRIS PILATUS
 * detectors. I only support images using the MIME mini-header and the
 * "none" or "byte_offset" compression algorithms. I do not support the
 * "unsigned 1-bit integer" and "signed 32-bit complex IEEE" data element
 * types.
 */
public class CbfImageReader extends ImageReader {
  private static final String QUOTE_CHARACTERS = "'\"";
  private static final int[] QUOTE_CODE_POINTS;
  private static final String PILATUS_SPACE_CHARACTERS = " #=:,()";
  private static final int[] PILATUS_SPACE_CODE_POINTS;
  private static final Pattern PILATUS_SPACE = Pattern.compile("[ #=:,()]");
  private static final Pattern PILATUS_SPACE_SPLIT = Pattern.compile(PILATUS_SPACE.pattern() + "+");
  private static final Pattern PILATUS_KEY_VALUE = Pattern.compile("^# +([a-zA-Z0-9_]+)" +
    PILATUS_SPACE.pattern() + "+(.*)");
  private static final Pattern PILATUS_DATE = Pattern.compile("^# +(\\d{4}[/-].+)");

  private static final String[] HEADER_CONVENTION_MAJORS = {
    "SLS_1.", "PILATUS_1."
  };
  private static final String[] HEADER_CONVENTIONS = {
    "SLS_1.0", "SLS_1.1", "SLS_1.2",
    "PILATUS_1.0", "PILATUS_1.1", "PILATUS_1.2"
  };

  static {
    QUOTE_CODE_POINTS = toCodePoints(QUOTE_CHARACTERS);
    Arrays.sort(QUOTE_CODE_POINTS);
    PILATUS_SPACE_CODE_POINTS = toCodePoints(PILATUS_SPACE_CHARACTERS);
    Arrays.sort(PILATUS_SPACE_CODE_POINTS);

    Arrays.sort(HEADER_CONVENTIONS);
  }

  private CbfMetadata metadata;

  public CbfImageReader(ImageReaderSpi originatingProvider) {
    super(originatingProvider);
  }

  @Override
  public int getNumImages(boolean allowSearch) throws IOException {
    return 1;
  }

  @Override
  public int getWidth(int imageIndex) throws IOException {
    checkIndex(imageIndex);
    return getMetadata().getDataWidth();
  }

  @Override
  public int getHeight(int imageIndex) throws IOException {
    checkIndex(imageIndex);
    return getMetadata().getDataHeight();
  }

  @Override
  public Iterator<ImageTypeSpecifier> getImageTypes(int imageIndex) throws IOException {
    checkIndex(imageIndex);
    List<ImageTypeSpecifier> result = new ArrayList<ImageTypeSpecifier>();
    CbfElementType elementType = getMetadata().getDataElementType();
    switch (elementType) {
      case UNSIGNED_8_BIT_INTEGER:
        result.add(ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_BYTE_GRAY));
        break;
      case SIGNED_8_BIT_INTEGER:
        /*
         * WORKAROUND: ImageTypeSpecifier.createGrayscale throws an
         * IllegalArgumentException if specify 8 bits and
         * DataBuffer.TYPE_BYTE. The documentation for the createGrayscale
         * method leads me to believe it should work, but it doesn't. So,
         * using 16 bits and DataBuffer.TYPE_SHORT instead.
         */
        result.add(ImageTypeSpecifier.createGrayscale(16, DataBuffer.TYPE_SHORT, true));
        break;
      case UNSIGNED_16_BIT_INTEGER:
        result.add(ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_USHORT_GRAY));
        break;
      case SIGNED_16_BIT_INTEGER:
        result.add(ImageTypeSpecifier.createGrayscale(16, DataBuffer.TYPE_SHORT, true));
        break;
      case UNSIGNED_32_BIT_INTEGER:
        result.add(ImageTypeSpecifier.createInterleaved(
          ColorSpace.getInstance(ColorSpace.CS_GRAY), new int[] { 0 }, DataBuffer.TYPE_FLOAT,
          false, false));
        result.add(ImageTypeSpecifier.createInterleaved(
          ColorSpace.getInstance(ColorSpace.CS_GRAY), new int[] { 0 }, DataBuffer.TYPE_INT, false,
          false));
        break;
      case SIGNED_32_BIT_INTEGER:
        result.add(ImageTypeSpecifier.createInterleaved(
          ColorSpace.getInstance(ColorSpace.CS_GRAY), new int[] { 0 }, DataBuffer.TYPE_INT, false,
          false));
        result.add(ImageTypeSpecifier.createInterleaved(
          ColorSpace.getInstance(ColorSpace.CS_GRAY), new int[] { 0 }, DataBuffer.TYPE_FLOAT,
          false, false));
        break;
      case SIGNED_32_BIT_REAL_IEEE:
        result.add(ImageTypeSpecifier.createInterleaved(
          ColorSpace.getInstance(ColorSpace.CS_GRAY), new int[] { 0 }, DataBuffer.TYPE_FLOAT,
          false, false));
        break;
      case SIGNED_64_BIT_REAL_IEEE:
        result.add(ImageTypeSpecifier.createInterleaved(
          ColorSpace.getInstance(ColorSpace.CS_GRAY), new int[] { 0 }, DataBuffer.TYPE_DOUBLE,
          false, false));
        break;
      default:
        throw new RuntimeException("Bug; element type: " + elementType);
    }
    return result.iterator();
  }

  @Override
  public IIOMetadata getStreamMetadata() throws IOException {
    return null;
  }

  @Override
  public IIOMetadata getImageMetadata(int imageIndex) throws IOException {
    checkIndex(imageIndex);
    return getMetadata();
  }

  @Override
  public ImageReadParam getDefaultReadParam() {
    return new CbfImageReadParam();
  }

  /**
   * Reads a CBF image.
   * <p>
   * The destination type for unsigned 32-bit integer source data is
   * <code>float</code> by default, but <code>int</code> is supported too.
   * When the destination type is <code>float</code>, some less significant
   * digits can be lost. When the destination type is <code>int</code>,
   * values larger than <code>Integer.MAX_VALUE</code> will be capped at
   * <code>Integer.MAX_VALUE</code> and a warning will be sent to registered
   * <code>IIOReadWarningListener</code> listeners.
   */
  /*
   * Optimized for speed. Reading the image data all at once into memory
   * results in a 25 times speed-up. Avoiding method invocations also speeds
   * things up which is why this method is quite long and has duplicate code
   * in the byte_offset decompression implementation.
   */
  @Override
  public BufferedImage read(int imageIndex, ImageReadParam irparam) throws IOException {
    checkIndex(imageIndex);
    clearAbortRequest();
    processImageStarted(0);

    ImageReadParam param = (irparam == null) ? getDefaultReadParam() : irparam;
    boolean minimizeMemoryUse = false;
    if (param instanceof CbfImageReadParam) {
      minimizeMemoryUse = ((CbfImageReadParam)param).shouldMinimizeMemoryUse();
    }
    int srcWidth = getWidth(0);
    int srcHeight = getHeight(0);

    BufferedImage dstImage = getDestination(param, getImageTypes(0), srcWidth, srcHeight);
    Rectangle srcRegion = new Rectangle();
    Rectangle dstRegion = new Rectangle();
    computeRegions(param, srcWidth, srcHeight, dstImage, srcRegion, dstRegion);
    int srcRegionMaxX = srcRegion.x + srcRegion.width - 1;
    int srcRegionMaxY = srcRegion.y + srcRegion.height - 1;
    int dstRegionMaxX = dstRegion.x + dstRegion.width - 1;

    int srcXSubsamp = param.getSourceXSubsampling();
    int srcYSubsamp = param.getSourceYSubsampling();

    int dstDataType = dstImage.getSampleModel().getDataType();
    checkReadParamBandSettings(param, 1, dstImage.getSampleModel().getNumBands());
    int[] dstBands = param.getDestinationBands();
    int dstBand = (dstBands == null) ? 0 : dstBands[0];
    WritableRaster dstRaster = dstImage.getRaster();

    CbfMetadata mdata = getMetadata();
    ImageInputStream stream = getNonNullInput();
    long streamStartOfData = stream.getStreamPosition();
    stream.mark();

    CbfData data;
    byte[] dataArray = null;
    int dataSize = mdata.getDataSize();

    MessageDigest md5Digest = null;
    byte[] expectedMd5 = mdata.getDataMd5();
    byte[] actualMd5 = null;
    if (expectedMd5 != null) {
      try {
        md5Digest = MessageDigest.getInstance("MD5");
      } catch (NoSuchAlgorithmException e) {
        processWarningOccurred("Failed to find MD5 algorithm for data MD5 digest calculation");
      }
    }

    if (minimizeMemoryUse || dataSize == 0) {
      data = new ImageInputStreamCbfData(stream);
    } else {
      dataArray = new byte[dataSize];
      stream.readFully(dataArray);
      data = new ByteArrayCbfData(dataArray);
    }

    if (dataArray != null && md5Digest != null) {
      md5Digest.update(dataArray, 0, dataArray.length);
      actualMd5 = md5Digest.digest();
      if (!MessageDigest.isEqual(expectedMd5, actualMd5)) {
        throw new IIOException("MIME header MD5 does not match data");
      }
    }

    data.order(mdata.getDataElementByteOrder());

    boolean readEntireImage = srcRegion.width == srcWidth && srcRegion.height == srcHeight;
    int progressRowPeriod = Math.max(1, (srcRegionMaxY + 1) / 100);
    CbfElementType elementType = mdata.getDataElementType();
    boolean unsigned = !elementType.isSigned();

    if (mdata.getDataCompression().equals(CbfCompression.NONE)) {
      long elementCount = 0;
      int dstY = dstRegion.y;
      int srcYSubsampCount = 0;
      for (int srcY = 0; srcY < srcHeight; srcY++) {
        if (abortRequested()) {
          processReadAborted();
          return dstImage;
        }
        if (srcY % progressRowPeriod == 0) {
          processImageProgress(((float)srcY / (float)(srcRegionMaxY + 1)) * 100.0f);
        }
        if (srcYSubsamp != 1 && srcY > srcRegion.y && srcY <= srcRegionMaxY) {
          srcYSubsampCount++;
          if (srcYSubsampCount == srcYSubsamp) srcYSubsampCount = 0;
        }
        int dstX = dstRegion.x;
        int srcXSubsampCount = 0;
        for (int srcX = 0; srcX < srcWidth; srcX++) {
          if (srcXSubsamp != 1 && srcX > srcRegion.y && srcX <= srcRegionMaxX) {
            srcXSubsampCount++;
            if (srcXSubsampCount == srcXSubsamp) srcXSubsampCount = 0;
          }
          boolean shouldSet =
            srcXSubsampCount == 0 && srcYSubsampCount == 0 && srcRegion.contains(srcX, srcY);
          if (!shouldSet && srcX > srcRegionMaxX && srcY > srcRegionMaxY) {
            srcY = srcHeight;
            srcX = srcWidth;
            break;
          }
          switch (elementType) {
            case UNSIGNED_8_BIT_INTEGER:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getUbyte());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(1);
              }
              break;
            case SIGNED_8_BIT_INTEGER:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getByte());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(1);
              }
              break;
            case UNSIGNED_16_BIT_INTEGER:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getUshort());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(2);
              }
              break;
            case SIGNED_16_BIT_INTEGER:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getShort());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(2);
              }
              break;
            case UNSIGNED_32_BIT_INTEGER:
              if (shouldSet) {
                long value = data.getUint();
                if (dstDataType == DataBuffer.TYPE_FLOAT) {
                  dstRaster.setSample(dstX, dstY, dstBand, (float)value);
                } else {
                  if (value <= Integer.MAX_VALUE) {
                    dstRaster.setSample(dstX, dstY, dstBand, (int)value);
                  } else {
                    dstRaster.setSample(dstX, dstY, dstBand, Integer.MAX_VALUE);
                    processWarningOccurred("Decoded pixel value " + value + " capped to "
                      + Integer.MAX_VALUE + " to fit in signed 32-bit integer range");
                  }
                }
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(4);
              }
              break;
            case SIGNED_32_BIT_INTEGER:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getInt());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(4);
              }
              break;
            case SIGNED_32_BIT_REAL_IEEE:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getFloat());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(4);
              }
              break;
            case SIGNED_64_BIT_REAL_IEEE:
              if (shouldSet) {
                dstRaster.setSample(dstX, dstY, dstBand, data.getDouble());
                dstX++;
                if (dstX > dstRegionMaxX) dstY++;
              } else {
                data.skipBytes(8);
              }
              break;
            default:
              throw new RuntimeException("Bug; element type: " + elementType);
          }
          elementCount++;
        }
      }
      if (readEntireImage && elementCount != mdata.getDataNumElements()) {
        throw new IIOException("Expected " + mdata.getDataNumElements()
          + " element count, but was " + elementCount);
      }
    } else if (mdata.getDataCompression().equals(CbfCompression.BYTE_OFFSET)) {
      long delta = 0;
      long value = 0;
      long elementCount = 0;
      int dstY = dstRegion.y;
      int srcYSubsampCount = 0;
      for (int srcY = 0; srcY < srcHeight; srcY++) {
        if (abortRequested()) {
          processReadAborted();
          return dstImage;
        }
        if (srcY % progressRowPeriod == 0) {
          processImageProgress(((float)srcY / (float)(srcRegionMaxY + 1)) * 100.0f);
        }
        if (srcYSubsamp != 1 && srcY > srcRegion.y && srcY <= srcRegionMaxY) {
          srcYSubsampCount++;
          if (srcYSubsampCount == srcYSubsamp) srcYSubsampCount = 0;
        }
        int dstX = dstRegion.x;
        int srcXSubsampCount = 0;
        for (int srcX = 0; srcX < srcWidth; srcX++) {
          if (srcXSubsamp != 1 && srcX > srcRegion.x && srcX <= srcRegionMaxX) {
            srcXSubsampCount++;
            if (srcXSubsampCount == srcXSubsamp) srcXSubsampCount = 0;
          }
          boolean shouldSet =
            srcXSubsampCount == 0 && srcYSubsampCount == 0 && srcRegion.contains(srcX, srcY);
          if (!shouldSet && srcX > srcRegionMaxX && srcY > srcRegionMaxY) {
            srcX = srcWidth;
            srcY = srcHeight;
            break;
          }
          delta = data.getByte();
          if (delta != Byte.MIN_VALUE) {
            value += delta;
            if (unsigned) value &= 0xFFFFFFFFL;
            if (shouldSet) {
              if (CbfElementType.UNSIGNED_32_BIT_INTEGER.equals(elementType)) {
                if (dstDataType == DataBuffer.TYPE_FLOAT) {
                  dstRaster.setSample(dstX, dstY, dstBand, (float)value);
                } else {
                  if (value <= Integer.MAX_VALUE) {
                    dstRaster.setSample(dstX, dstY, dstBand, (int)value);
                  } else {
                    dstRaster.setSample(dstX, dstY, dstBand, Integer.MAX_VALUE);
                    processWarningOccurred("Decoded pixel value " + value + " capped to "
                      + Integer.MAX_VALUE + " to fit in signed 32-bit integer range");
                  }
                }
              } else {
                dstRaster.setSample(dstX, dstY, dstBand, (int)value);
              }
              dstX++;
              if (dstX > dstRegionMaxX) dstY++;
            }
            elementCount++;
            continue;
          }
          delta = data.getShort();
          if (delta != Short.MIN_VALUE) {
            value += delta;
            if (unsigned) value &= 0xFFFFFFFFL;
            if (shouldSet) {
              if (CbfElementType.UNSIGNED_32_BIT_INTEGER.equals(elementType)) {
                if (dstDataType == DataBuffer.TYPE_FLOAT) {
                  dstRaster.setSample(dstX, dstY, dstBand, (float)value);
                } else {
                  if (value <= Integer.MAX_VALUE) {
                    dstRaster.setSample(dstX, dstY, dstBand, (int)value);
                  } else {
                    dstRaster.setSample(dstX, dstY, dstBand, Integer.MAX_VALUE);
                    processWarningOccurred("Decoded pixel value " + value + " capped to "
                      + Integer.MAX_VALUE + " to fit in signed 32-bit integer range");
                  }
                }
              } else {
                dstRaster.setSample(dstX, dstY, dstBand, (int)value);
              }
              dstX++;
              if (dstX > dstRegionMaxX) dstY++;
            }
            elementCount++;
            continue;
          }
          delta = data.getInt();
          if (delta != Integer.MIN_VALUE) {
            value += delta;
            if (unsigned) value &= 0xFFFFFFFFL;
            if (shouldSet) {
              if (CbfElementType.UNSIGNED_32_BIT_INTEGER.equals(elementType)) {
                if (dstDataType == DataBuffer.TYPE_FLOAT) {
                  dstRaster.setSample(dstX, dstY, dstBand, (float)value);
                } else {
                  if (value <= Integer.MAX_VALUE) {
                    dstRaster.setSample(dstX, dstY, dstBand, (int)value);
                  } else {
                    dstRaster.setSample(dstX, dstY, dstBand, Integer.MAX_VALUE);
                    processWarningOccurred("Decoded pixel value " + value + " capped to "
                      + Integer.MAX_VALUE + " to fit in signed 32-bit integer range");
                  }
                }
              } else {
                dstRaster.setSample(dstX, dstY, dstBand, (int)value);
              }
              dstX++;
              if (dstX > dstRegionMaxX) dstY++;
            }
            elementCount++;
            continue;
          }
          delta = data.getLong();
          value += delta;
          if (unsigned) value &= 0xFFFFFFFFL;
          if (shouldSet) {
            if (CbfElementType.UNSIGNED_32_BIT_INTEGER.equals(elementType)) {
              if (dstDataType == DataBuffer.TYPE_FLOAT) {
                dstRaster.setSample(dstX, dstY, dstBand, (float)value);
              } else {
                if (value <= Integer.MAX_VALUE) {
                  dstRaster.setSample(dstX, dstY, dstBand, (int)value);
                } else {
                  dstRaster.setSample(dstX, dstY, dstBand, Integer.MAX_VALUE);
                  processWarningOccurred("Decoded pixel value " + value + " capped to "
                    + Integer.MAX_VALUE + " to fit in signed 32-bit integer range");
                }
              }
            } else {
              dstRaster.setSample(dstX, dstY, dstBand, (int)value);
            }
            dstX++;
            if (dstX > dstRegionMaxX) dstY++;
          }
          elementCount++;
        }
      }
      if (readEntireImage && elementCount != mdata.getDataNumElements()) {
        throw new IIOException("Expected " + mdata.getDataNumElements()
          + " element count, but was " + elementCount);
      }
    } else {
      throw new RuntimeException("Bug; compression: " + mdata.getDataCompression());
    }

    if (abortRequested()) {
      processReadAborted();
      return dstImage;
    }

    if (readEntireImage && md5Digest != null && dataArray == null) {
      long streamEndOfData = stream.getStreamPosition();
      if (streamEndOfData - streamStartOfData <= Integer.MAX_VALUE) {
        dataSize = (int)(streamEndOfData - streamStartOfData);
        if (minimizeMemoryUse) {
          for (int i = 0; i < dataSize; i++)
            md5Digest.update(stream.readByte());
        } else {
          dataArray = new byte[dataSize];
          stream.reset();
          stream.readFully(dataArray);
          md5Digest.update(dataArray, 0, dataArray.length);
        }
        actualMd5 = md5Digest.digest();
        if (!MessageDigest.isEqual(expectedMd5, actualMd5)) {
          throw new IIOException("MIME header MD5 does not match data");
        }
      }
    }

    processImageComplete();

    return dstImage;
  }

  @Override
  public void setInput(Object input, boolean seekForwardOnly, boolean ignoreMetadata) {
    super.setInput(input, seekForwardOnly, ignoreMetadata);
    metadata = null;
  }

  private void checkIndex(int imageIndex) {
    if (imageIndex != 0) throw new IndexOutOfBoundsException("imageIndex != 0");
  }

  private ImageInputStream getNonNullInput() {
    Object input = getInput();
    if (input == null) throw new IllegalStateException("No input stream");
    return (ImageInputStream)input;
  }

  private CbfMetadata getMetadata() throws IIOException {
    if (metadata == null) readMetadata();
    return metadata;
  }

  private void readMetadata() throws IIOException {
    ImageInputStream stream = getNonNullInput();
    CbfMetadata mdata = new CbfMetadata();
    readMetadataIdentifier(stream, mdata);
    readMetadataDataBlockName(stream, mdata);
    readMetadataHeader(stream, mdata);
    readMetadataMimeMiniHeader(stream, mdata);
    metadata = mdata;
  }

  private void readMetadataHeader(ImageInputStream stream, CbfMetadata mdata)
      throws IIOException {
    try {
      String line = readNonemptyNoncommentLine(stream);
      if (line.startsWith("_array_data.header_convention ")) {
        String convention = trimQuotes(line.substring("_array_data.header_convention ".length()));
        boolean foundMatchingHeaderConventionMajor = false;
        for (String each : HEADER_CONVENTION_MAJORS) {
          if (convention.startsWith(each)) {
            foundMatchingHeaderConventionMajor = true;
            break;
          }
        }
        if (!foundMatchingHeaderConventionMajor) {
          throw new IIOException("Unrecognized header convention: " + convention);
        }
        mdata.setHeaderConvention(convention);
        if (Arrays.binarySearch(HEADER_CONVENTIONS, convention) < 0) {
          processWarningOccurred("Unrecognized header convention: " + convention);
        }
        line = readNonemptyNoncommentLine(stream);
      } else {
        processWarningOccurred("No header convention found");
      }
      if (!line.equals("_array_data.header_contents")) {
        throw new IIOException("Expected _array_data.header_contents tag");
      }
      line = readNonemptyNoncommentLine(stream);
      if (!line.equals(";")) {
        throw new IIOException("Expected ';' line after _array_data.header_contents");
      }

      for (;;) {
        line = readNonemptyLine(stream);
        if (line.equals(";")) break;
        Matcher matcher = PILATUS_KEY_VALUE.matcher(line);
        if (matcher.matches()) {
          String key = matcher.group(1);
          String value = matcher.group(2);
          if (key.equals("Detector")) {
            mdata.setHeaderDetector(parseHeaderDetector(value));
          } else if (key.equals("Pixel_size")) {
            mdata.setHeaderPixelSize(parseHeaderPixelSize(value));
          } else if (key.equals("Silicon")) {
            Object[] typeAndThickness = parseHeaderSensor(value);
            mdata.setHeaderSensorType((String)typeAndThickness[0]);
            mdata.setHeaderSensorThickness((Float)typeAndThickness[1]);
          } else if (key.equals("Exposure_time")) {
            mdata.setHeaderExposureTime(parseHeaderExposureTime(value));
          } else if (key.equals("Exposure_period")) {
            mdata.setHeaderExposurePeriod(parseHeaderExposurePeriod(value));
          } else if (key.equals("Tau")) {
            mdata.setHeaderTau(parseHeaderTau(value));
          } else if (key.equals("Count_cutoff")) {
            mdata.setHeaderCountCutoff(parseHeaderCountCutoff(value));
          } else if (key.equals("Threshold_setting")) {
            mdata.setHeaderThreshold(parseHeaderThreshold(value));
          } else if (key.equals("Gain_setting")) {
            Object[] typeAndVrf = parseHeaderGain(value);
            mdata.setHeaderGainType((String)typeAndVrf[0]);
            mdata.setHeaderGainVrf((Float)typeAndVrf[1]);
          } else if (key.equals("N_excluded_pixels")) {
            mdata.setHeaderNumExcludedPixels(parseHeaderNumExcludedPixels(value));
          } else if (key.equals("Excluded_pixels")) {
            mdata.setHeaderExcludedPixels(parseHeaderExcludedPixels(value));
          } else if (key.equals("Flat_field")) {
            mdata.setHeaderFlatField(parseHeaderFlatField(value));
          } else if (key.equals("Trim_directory")) {
            mdata.setHeaderTrimDirectory(parseHeaderTrimDirectory(value));
          } else if (key.equals("Trim_file")) {
            mdata.setHeaderTrimFile(parseHeaderTrimFile(value));
          } else if (key.equals("Image_path")) {
            mdata.setHeaderImagePath(parseHeaderImagePath(value));
          } else if (key.equals("Wavelength")) {
            mdata.setHeaderWavelength(parseHeaderWavelength(value));
          } else if (key.equals("Energy_range")) {
            mdata.setHeaderEnergyRange(parseHeaderEnergyRange(value));
          } else if (key.equals("Detector_distance")) {
            mdata.setHeaderDetectorDistance(parseHeaderDetectorDistance(value));
          } else if (key.equals("Detector_Voffset")) {
            mdata.setHeaderDetectorVOffset(parseHeaderDetectorVOffset(value));
          } else if (key.equals("Beam_xy")) {
            mdata.setHeaderBeamXY(parseHeaderBeamXY(value));
          } else if (key.equals("Flux")) {
            try {
              mdata.setHeaderFlux(parseHeaderFlux(value));
            } catch (IIOException e) {
              /* IGNORED */
            }
            mdata.setHeaderFluxString(parseHeaderFluxString(value));
          } else if (key.equals("Filter_transmission")) {
            mdata.setHeaderFilterTransmission(parseHeaderFilterTransmission(value));
          } else if (key.equals("Start_angle")) {
            mdata.setHeaderStartAngle(parseHeaderStartAngle(value));
          } else if (key.equals("Angle_increment")) {
            mdata.setHeaderAngleIncrement(parseHeaderAngleIncrement(value));
          } else if (key.equals("Detector_2theta")) {
            mdata.setHeaderDetector2Theta(parseHeaderDetector2Theta(value));
          } else if (key.equals("Polarization")) {
            mdata.setHeaderPolarization(parseHeaderPolarization(value));
          } else if (key.equals("Alpha")) {
            mdata.setHeaderAlpha(parseHeaderAlpha(value));
          } else if (key.equals("Kappa")) {
            mdata.setHeaderKappa(parseHeaderKappa(value));
          } else if (key.equals("Phi")) {
            mdata.setHeaderPhi(parseHeaderPhi(value));
          } else if (key.equals("Phi_increment")) {
            mdata.setHeaderPhiIncrement(parseHeaderPhiIncrement(value));
          } else if (key.equals("Chi")) {
            mdata.setHeaderChi(parseHeaderChi(value));
          } else if (key.equals("Chi_increment")) {
            mdata.setHeaderChiIncrement(parseHeaderChiIncrement(value));
          } else if (key.equals("Omega")) {
            mdata.setHeaderOmega(parseHeaderOmega(value));
          } else if (key.equals("Omega_increment")) {
            mdata.setHeaderOmegaIncrement(parseHeaderOmegaIncrement(value));
          } else if (key.equals("Oscillation_axis")) {
            mdata.setHeaderOscillationAxis(parseHeaderOscillationAxis(value));
          } else if (key.equals("N_oscillations")) {
            mdata.setHeaderNumOscillations(parseHeaderNumOscillations(value));
          } else if (key.equals("Start_position")) {
            mdata.setHeaderStartPosition(parseHeaderStartPosition(value));
          } else if (key.equals("Position_increment")) {
            mdata.setHeaderPositionIncrement(parseHeaderPositionIncrement(value));
          } else if (key.equals("Shutter_time")) {
            mdata.setHeaderShutterTime(parseHeaderShutterTime(value));
          }
        } else {
          matcher = PILATUS_DATE.matcher(line);
          if (matcher.matches()) {
            mdata.setHeaderDate(parseHeaderDate(matcher.group(1)));
          } else {
            mdata.addHeaderUnrecognizedLine(line);
            processWarningOccurred("Unrecognized header line: " + line);
          }
        }
      }
    } catch (IOException e) {
      throw new IIOException("Error reading header", e);
    }
  }

  private void readMetadataMimeMiniHeader(ImageInputStream stream, CbfMetadata mdata)
      throws IIOException {
    try {
      String line = readNonemptyNoncommentLine(stream);
      if (!line.equals("_array_data.data")) {
        throw new IIOException("Expected _array_data.data tag");
      }
      line = readNonemptyNoncommentLine(stream);
      if (!line.equals(";")) {
        throw new IIOException("Expected ';' line after _array_data.data");
      }
      line = readNonemptyNoncommentLine(stream);
      if (!line.equals("--CIF-BINARY-FORMAT-SECTION--")) {
        throw new IIOException("Expected --CIF-BINARY-FORMAT-SECTION--");
      }

      String nextLine = stream.readLine();
      for (;;) {
        line = nextLine;
        if (line.length() == 0) break;
        nextLine = stream.readLine();
        if (nextLine.startsWith(" ") || nextLine.startsWith("\t")) {
          nextLine = line + nextLine;
        } else if (line.startsWith("Content-Type:")) {
          String value = line.substring("Content-Type:".length());
          Object[] contentType = parseMimeMiniHeaderContentType(value);
          CbfCompression comp = (CbfCompression)contentType[1];
          CbfElementType type = mdata.getDataElementType();
          if (type != null && !comp.compatibleWith(type)) {
            throw new IIOException(comp + " compression incompatible with element type " + type);
          }
          mdata.setDataCompression(comp);
        } else if (line.startsWith("Content-Transfer-Encoding:")) {
          String value = line.substring("Content-Transfer-Encoding:".length());
          parseMimeMiniHeaderContentTransferEncoding(value);
        } else if (line.startsWith("X-Binary-Size:")) {
          String value = line.substring("X-Binary-Size:".length());
          mdata.setDataSize(parseMimeMiniHeaderBinarySize(value));
        } else if (line.startsWith("X-Binary-ID:")) {
          String value = line.substring("X-Binary-ID:".length());
          mdata.setDataId(parseMimeMiniHeaderBinaryId(value));
        } else if (line.startsWith("X-Binary-Element-Type:")) {
          String value = line.substring("X-Binary-Element-Type:".length());
          CbfElementType type = parseMimeMiniHeaderBinaryElementType(value);
          CbfCompression comp = mdata.getDataCompression();
          if (comp != null && !comp.compatibleWith(type)) {
            throw new IIOException(comp + " compression incompatible with element type " + type);
          }
          mdata.setDataElementType(type);
        } else if (line.startsWith("X-Binary-Element-Byte-Order:")) {
          String value = line.substring("X-Binary-Element-Byte-Order:".length());
          mdata.setDataElementByteOrder(parseMimeMiniHeaderBinaryElementByteOrder(value));
        } else if (line.startsWith("Content-MD5:")) {
          String value = line.substring("Content-MD5:".length());
          mdata.setDataMd5(parseMimeMiniHeaderContentMd5(value));
        } else if (line.startsWith("X-Binary-Number-of-Elements:")) {
          String value = line.substring("X-Binary-Number-of-Elements:".length());
          mdata.setDataNumElements(parseMimeMiniHeaderBinaryNumberOfElements(value));
        } else if (line.startsWith("X-Binary-Size-Fastest-Dimension:")) {
          String value = line.substring("X-Binary-Size-Fastest-Dimension:".length());
          mdata.setDataWidth(parseMimeMiniHeaderBinarySizeFastestDimension(value));
        } else if (line.startsWith("X-Binary-Size-Second-Dimension:")) {
          String value = line.substring("X-Binary-Size-Second-Dimension:".length());
          mdata.setDataHeight(parseMimeMiniHeaderBinarySizeSecondDimension(value));
        } else if (line.startsWith("X-Binary-Size-Padding:")) {
          String value = line.substring("X-Binary-Size-Padding:".length());
          mdata.setDataPadding(parseMimeMiniHeaderBinarySizePadding(value));
        } else {
          processWarningOccurred("Unrecognized MIME mini-header line: " + line);
        }
      }

      byte[] octets = { 0x0C, 0x1A, 0x04, (byte)0xD5 };
      for (int i = 0; i < octets.length; i++) {
        byte octet = stream.readByte();
        if (octet != octets[i]) {
          throw new IIOException(String.format(
            "Expected %#x in octet series before binary data, but was %#x", octets[i], octet));
        }
      }
    } catch (IOException e) {
      throw new IIOException("Error reading MIME mini-header", e);
    }
  }

  private static void readMetadataIdentifier(ImageInputStream stream, CbfMetadata mdata)
      throws IIOException {
    String identifier;
    try {
      identifier = stream.readLine();
    } catch (IOException e) {
      throw new IIOException("Error reading identifier", e);
    }
    if (!identifier.startsWith("###CBF: VERSION")) throw new IIOException("Not a CBF file");
    mdata.setIdentifier(identifier);
  }

  private static void readMetadataDataBlockName(ImageInputStream stream, CbfMetadata mdata)
      throws IIOException {
    try {
      String line = readNonemptyNoncommentLine(stream);
      if (!line.startsWith("data_")) throw new IIOException("Expected data block declaration");
      mdata.setDataBlockName(line.substring("data_".length()));
    } catch (IOException e) {
      throw new IIOException("Error reading data block name", e);
    }
  }

  private static String parseHeaderDetector(String value) {
    return trimPilatusSpace(value.trim());
  }

  private static Date parseHeaderDate(String value) throws IIOException {
    String dateText = trimPilatusSpace(value.trim());
    String[] formatTexts = { "yyyy-MMM-dd'T'HH:mm:ss.SSS", "yyyy-MM-dd'T'HH:mm:ss.SSS",
      "yyyy/MMM/dd HH:mm:ss.SSS", "yyyy/MM/dd HH:mm:ss.SSS" };
    for (String each : formatTexts) {
      DateFormat format = new SimpleDateFormat(each, Locale.US);
      ParsePosition position = new ParsePosition(0);
      Date result = format.parse(dateText, position);
      if (result != null && position.getIndex() >= dateText.length()) return result;
    }
    throw new IIOException("Error parsing header date: " + dateText);
  }

  private static float[] parseHeaderPixelSize(String value) throws IIOException {
    return parseHeaderFloatArray(value, new int[] {0, 3}, "pixel size");
  }

  private static Object[] parseHeaderSensor(String value) throws IIOException {
    return new Object[] {"Silicon", parseHeaderFloat(value, 2, "sensor thickness")};
  }

  private static Float parseHeaderExposureTime(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "exposure time");
  }

  private static Float parseHeaderExposurePeriod(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "exposure period");
  }

  private static Float parseHeaderTau(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "tau");
  }

  private static Integer parseHeaderCountCutoff(String value) throws IIOException {
    return parseHeaderInteger(value, 0, "count cutoff");
  }

  private static Float parseHeaderThreshold(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "threshold");
  }

  private static Object[] parseHeaderGain(String value) throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    String[] parts = PILATUS_SPACE_SPLIT.split(trimmedValue);
    if (parts.length < 1) {
      throw new IIOException("Error parsing header gain: " + trimmedValue);
    }
    Object[] result = new Object[] {trimmedValue, null};
    if (parts.length == 1) return result;
    int vrfIndex = Arrays.asList(parts).indexOf("vrf");
    if (vrfIndex == -1 || vrfIndex == parts.length - 1) return result;
    StringBuilder typeBuilder = new StringBuilder();
    for (int i = 0; i < vrfIndex; i++) {
      if (i != 0) typeBuilder.append(' ');
      typeBuilder.append(parts[i]);
    }
    String type = typeBuilder.toString();
    if (type.equals("") || type.equals("not implemented")) {
      type = null;
    }
    result[0] = type;
    int vrfValueIndex = vrfIndex + 1;
    Float vrf;
    try {
      vrf = Float.valueOf(parts[vrfValueIndex]);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing header gain vrf: " + parts[vrfValueIndex], e);
    }
    result[1] = vrf;
    return result;
  }

  private static Integer parseHeaderNumExcludedPixels(String value) throws IIOException {
    return parseHeaderInteger(value, 0, "number of excluded pixels");
  }

  private static String parseHeaderExcludedPixels(String value) throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    return (trimmedValue.length() == 0 || trimmedValue.equals("nil")) ? null : trimmedValue;
  }

  private static String parseHeaderFlatField(String value) throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    return (trimmedValue.length() == 0 || trimmedValue.equals("nil")) ? null : trimmedValue;
  }

  private static String parseHeaderTrimDirectory(String value) throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    return (trimmedValue.length() == 0 || trimmedValue.equals("nil")) ? null : trimmedValue;
  }

  private static String parseHeaderTrimFile(String value) throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    return (trimmedValue.length() == 0 || trimmedValue.equals("nil")) ? null : trimmedValue;
  }

  private static String parseHeaderImagePath(String value) throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    return (trimmedValue.length() == 0 || trimmedValue.equals("nil")) ? null : trimmedValue;
  }

  private static Float parseHeaderWavelength(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "wavelength");
  }

  private static float[] parseHeaderEnergyRange(String value) throws IIOException {
    return parseHeaderFloatArray(value, new int[] {0, 1}, "energy range");
  }

  private static Float parseHeaderDetectorDistance(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "detector distance");
  }

  private static Float parseHeaderDetectorVOffset(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "vertical offset");
  }

  private static float[] parseHeaderBeamXY(String value) throws IIOException {
    return parseHeaderFloatArray(value, new int[] {0, 1}, "beam XY");
  }

  private static Float parseHeaderFlux(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "flux");
  }

  private static String parseHeaderFluxString(String value) throws IIOException {
    return trimPilatusSpace(value.trim());
  }

  private static Float parseHeaderFilterTransmission(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "filter transmission");
  }

  private static Float parseHeaderStartAngle(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "start angle");
  }

  private static Float parseHeaderAngleIncrement(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "angle increment");
  }

  private static Float parseHeaderDetector2Theta(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "detector 2theta");
  }

  private static Float parseHeaderPolarization(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "polarization");
  }

  private static Float parseHeaderAlpha(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "alpha");
  }

  private static Float parseHeaderKappa(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "kappa");
  }

  private static Float parseHeaderPhi(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "phi");
  }

  private static Float parseHeaderPhiIncrement(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "phi increment");
  }

  private static Float parseHeaderChi(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "chi");
  }

  private static Float parseHeaderChiIncrement(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "chi increment");
  }

  private static Float parseHeaderOmega(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "omega");
  }

  private static Float parseHeaderOmegaIncrement(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "omega increment");
  }

  private static String parseHeaderOscillationAxis(String value) {
    return trimPilatusSpace(value.trim());
  }

  private static Integer parseHeaderNumOscillations(String value) throws IIOException {
    return parseHeaderInteger(value, 0, "number of oscillations");
  }

  private static Float parseHeaderStartPosition(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "start position");
  }

  private static Float parseHeaderPositionIncrement(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "position increment");
  }

  private static Float parseHeaderShutterTime(String value) throws IIOException {
    return parseHeaderFloat(value, 0, "shutter time");
  }

  private static Float parseHeaderFloat(String value, int position, String label)
      throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    String[] parts = PILATUS_SPACE_SPLIT.split(trimmedValue);
    if (parts.length <= position) {
      throw new IIOException("Error parsing header " + label + ": " + trimmedValue);
    }
    try {
      return Float.valueOf(parts[position]);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing header " + label + ": " + parts[position], e);
    }
  }

  private static Integer parseHeaderInteger(String value, int position, String label)
      throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    String[] parts = PILATUS_SPACE_SPLIT.split(trimmedValue);
    if (parts.length <= position) {
      throw new IIOException("Error parsing header " + label + ": " + trimmedValue);
    }
    try {
      return Integer.valueOf(parts[position]);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing header " + label + ": " + parts[position], e);
    }
  }

  private static float[] parseHeaderFloatArray(String value, int[] positions, String label)
      throws IIOException {
    String trimmedValue = trimPilatusSpace(value.trim());
    String[] parts = PILATUS_SPACE_SPLIT.split(trimmedValue);
    if (parts.length <= positions[positions.length - 1]) {
      throw new IIOException("Error parsing header " + label + ": " + trimmedValue);
    }
    float[] result = new float[positions.length];
    for (int i = 0; i < result.length; i++) {
      try {
        result[i] = Float.parseFloat(parts[positions[i]]);
      } catch (NumberFormatException e) {
        throw new IIOException("Error parsing header " + label + ": " + parts[positions[i]], e);
      }
    }
    return result;
  }

  private static Object[] parseMimeMiniHeaderContentType(String value) throws IIOException {
    int semicolonIndex = value.indexOf(';');
    if (semicolonIndex < 0) semicolonIndex = value.length();
    String mediaType = value.substring(0, semicolonIndex).trim();
    if (!mediaType.equals("application/octet-stream")) {
      throw new IIOException("Unrecognized content type media type: " + mediaType);
    }
    int conversionsIndex = value.indexOf("conversions=", semicolonIndex + 1);
    if (conversionsIndex < 0) {
      throw new IIOException("Expected content type conversions; default x-CBF_PACKED unsupported");
    }
    String conversions = value.substring(conversionsIndex + "conversions=".length()).trim();
    if (conversions.startsWith("\"")) conversions = conversions.substring(1);
    int doubleQuoteIndex = conversions.indexOf('"');
    if (doubleQuoteIndex >= 0) conversions = conversions.substring(0, doubleQuoteIndex);
    try {
      return new Object[] { mediaType, CbfCompression.parse(conversions) };
    } catch (ParseException e) {
      throw new IIOException("Unsupported content type conversions: " + conversions, e);
    }
  }

  private static void parseMimeMiniHeaderContentTransferEncoding(String value) throws IIOException {
    String encoding = value.trim();
    if (encoding.equals("BINARY")) return;
    throw new IIOException("Unexpected encoding: " + encoding);
  }

  private static int parseMimeMiniHeaderBinarySize(String value) throws IIOException {
    String size = value.trim();
    try {
      return Integer.parseInt(size);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing binary size: " + size);
    }
  }

  private static int parseMimeMiniHeaderBinaryId(String value) throws IIOException {
    String id = value.trim();
    try {
      return Integer.parseInt(id);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing binary ID: " + id);
    }
  }

  private static CbfElementType parseMimeMiniHeaderBinaryElementType(String value)
      throws IIOException {
    String type = trimQuotes(value.trim());
    try {
      return CbfElementType.parse(type);
    } catch (ParseException e) {
      throw new IIOException("Unsupported binary element type: " + type, e);
    }
  }

  private static ByteOrder parseMimeMiniHeaderBinaryElementByteOrder(String value)
      throws IIOException {
    String order = value.trim();
    if (order.equals("LITTLE_ENDIAN")) {
      return ByteOrder.LITTLE_ENDIAN;
    } else if (order.equals("BIG_ENDIAN")) {
      return ByteOrder.BIG_ENDIAN;
    } else {
      throw new IIOException("Error parsing binary element byte order: " + order);
    }
  }

  private static byte[] parseMimeMiniHeaderContentMd5(String value) throws IIOException {
    String md5 = value.trim();
    try {
      return Base64.decode(md5);
    } catch (IllegalArgumentException e) {
      throw new IIOException("Error decoding MD5 digest: " + md5, e);
    }
  }

  private static int parseMimeMiniHeaderBinaryNumberOfElements(String value) throws IIOException {
    String count = value.trim();
    try {
      return Integer.parseInt(count);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing binary number of elements: " + count, e);
    }
  }

  private static int parseMimeMiniHeaderBinarySizeFastestDimension(String value)
      throws IIOException {
    String size = value.trim();
    try {
      return Integer.parseInt(size);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing binary size fastest dimension: " + size, e);
    }
  }

  private static int parseMimeMiniHeaderBinarySizeSecondDimension(String value)
      throws IIOException {
    String size = value.trim();
    try {
      return Integer.parseInt(size);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing binary size second dimension: " + size, e);
    }
  }

  private static int parseMimeMiniHeaderBinarySizePadding(String value) throws IIOException {
    String padding = value.trim();
    try {
      return Integer.parseInt(padding);
    } catch (NumberFormatException e) {
      throw new IIOException("Error parsing binary size padding: " + padding, e);
    }
  }

  private static String readNonemptyLine(ImageInputStream stream) throws IOException {
    String line;
    for (;;) {
      line = stream.readLine().trim();
      if (line.length() != 0) return line;
    }
  }

  private static String readNonemptyNoncommentLine(ImageInputStream stream) throws IOException {
    String line;
    for (;;) {
      line = stream.readLine().trim();
      if (line.length() != 0 && !line.startsWith("#")) return line;
    }
  }

  private static String trimQuotes(String s) {
    return JavaUtilities.trim(s, QUOTE_CODE_POINTS);
  }

  private static String trimPilatusSpace(String s) {
    return JavaUtilities.trim(s, PILATUS_SPACE_CODE_POINTS);
  }

  private static int[] toCodePoints(String s) {
    return JavaUtilities.toCodePoints(s);
  }
}
